"""Testing generate_errors module."""

import pytest
import numpy as np
import pandas as pd
from de import generate_errors


def test_negative_dynamic():
    ts_idx = pd.date_range(start="1/1/2019", periods=5)
    arr = np.array([1, 2, 3, 4, 5])
    ts = pd.DataFrame(index=ts_idx, data=arr)
    ts_nd = generate_errors.negative_dynamic(ts)
    assert ts_nd.iloc[0, 0] > 1
    assert ts_nd.iloc[4, 0] < 5


def test_positive_dynamic():
    ts_idx = pd.date_range(start="1/1/2019", periods=5)
    arr = np.array([1, 2, 3, 4, 5])
    ts = pd.DataFrame(index=ts_idx, data=arr)
    ts_pd = generate_errors.positive_dynamic(ts)
    assert ts_pd.iloc[0, 0] < 1
    assert ts_pd.iloc[4, 0] > 5


def test_positive_constant():
    arr = np.array([1, 2, 3, 4, 5])
    arr_pos = generate_errors.constant(arr)
    assert arr_pos[0] > 1
    assert arr_pos[-1] > 5


def test_negative_constant():
    arr = np.array([1, 2, 3, 4, 5])
    arr_pos = generate_errors.constant(arr, offset=0.5)
    assert arr_pos[0] < 1
    assert arr_pos[-1] < 5


def test_timing_shift():
    arr = np.array([1, 2, 3, 4, 5])
    date_rng = pd.date_range(start="1/1/2018", periods=5)
    ts = pd.DataFrame(index=date_rng, columns=["Q"], data=arr)
    ts_tim = generate_errors.timing(ts, shuffle=False)
    assert ts_tim["Q"].values[0] == 3
    assert ts_tim["Q"].values[-1] == 2


def test_timing_shuffling():
    arr = np.array([1, 2, 3, 4, 5])
    date_rng = pd.date_range(start="1/1/2018", periods=5)
    ts = pd.DataFrame(index=date_rng, columns=["Q"], data=arr)
    ts_tim = generate_errors.timing(ts)
    arr = np.array([1, 2, 3, 4, 5])
    diff = np.all(arr == ts_tim["Q"].values)
    assert diff == False



"""Testing nse module."""

import pytest
import numpy as np
import pandas as pd
from de import nse


def test_nse_for_arrays():
    eff = nse.calc_nse(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert eff == pytest.approx(0.5648252536640361, rel=1e-4)


def test_nse_dec_for_arrays():
    eff = nse.calc_nse_dec(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert eff == pytest.approx(0.9251076704923259, rel=1e-4)


def test_nse_beta_for_arrays():
    beta = nse.calc_nse_beta(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert beta == pytest.approx(0.2907828720750609, rel=1e-4)


def test_nse_alpha_for_arrays():
    alpha = nse.calc_nse_alpha(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert alpha == pytest.approx(1.2812057455166919, rel=1e-4)


def test_nse_r_for_arrays():
    lin_cor = nse.calc_nse_r(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert lin_cor == pytest.approx(0.8940281850583509, rel=1e-4)


def test_nse_for_equal_arrays():
    sig = nse.calc_nse(obs=np.array([1, 2, 3, 4, 5]), sim=np.array([1, 2, 3, 4, 5]))
    assert sig == 1


def test_nse_simulation_equals_obs_mean():
    sig = nse.calc_nse(obs=np.array([1, 2, 3, 4, 5]), sim=np.array([3, 3, 3, 3, 3]))
    assert sig == 0



"""Testing de module."""

import os
import sys

import pytest
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from de import de

baseline_dir = "images"

baseline_dir_single = os.path.join(baseline_dir, "de/single")
baseline_dir_multi = os.path.join(baseline_dir, "de/multi")

WIN = sys.platform.startswith("win")

# In some cases, the fonts on Windows can be quite different
DEFAULT_TOLERANCE = 10 if WIN else 2


def test_de_for_arrays():
    eff = de.calc_de(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert eff == pytest.approx(0.8202204384691575, rel=1e-4)


def test_lin_cor_for_arrays():
    r = de.calc_temp_cor(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert r == pytest.approx(0.8940281850583509, rel=1e-4)


def test_nonlin_cor_for_arrays():
    r = de.calc_temp_cor(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
        r="spearman",
    )
    assert r == pytest.approx(0.8406680016960504, rel=1e-4)


def test_bias_area_for_arrays():
    brel_res = de.calc_brel_res(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    b_area = de.calc_bias_area(brel_res)
    assert b_area == pytest.approx(0.1112908496732026, rel=1e-4)


def test_bias_dir_for_arrays():
    brel_res = de.calc_brel_res(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    b_dir = b_dir = de.calc_bias_dir(brel_res)
    assert b_dir == pytest.approx(-0.014705882352941155, rel=1e-4)


def test_bias_slope_for_arrays():
    brel_res = de.calc_brel_res(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    b_area = de.calc_bias_area(brel_res)
    b_dir = de.calc_bias_dir(brel_res)
    b_slope = de.calc_bias_slope(b_area, b_dir)
    assert b_slope == pytest.approx(0.1112908496732026, rel=1e-4)


def test_brel_mean_for_arrays():
    brel_mean = de.calc_brel_mean(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert brel_mean == pytest.approx(0.09330065359477124, rel=1e-4)


def test_brel_mean_simulation_equals_obs_mean():
    brel_mean = de.calc_brel_mean(
        obs=np.array([1, 2, 3, 4, 5]), sim=np.array([3, 3, 3, 3, 3])
    )
    assert brel_mean > 0.2


def test_brel_mean_for_equal_arrays():
    brel_mean = de.calc_brel_mean(obs=np.array([1, 2, 3]), sim=np.array([1, 2, 3]))
    assert brel_mean == 0


def test_brel_rest_simulation_equals_obs_mean():
    brel_res = de.calc_brel_res(
        obs=np.array([1, 2, 3, 4, 5]), sim=np.array([3, 3, 3, 3, 3])
    )
    b_area = de.calc_bias_area(brel_res)
    assert b_area == pytest.approx(0.5116666666666666, rel=1e-4)


def test_brel_rest_for_equal_arrays():
    brel_res = de.calc_brel_res(obs=np.array([1, 2, 3]), sim=np.array([1, 2, 3]))
    b_area = de.calc_bias_area(brel_res)
    assert b_area == 0


def test_bias_dir_simulation_equals_obs_mean():
    brel_res = de.calc_brel_res(
        obs=np.array([1, 2, 3, 4, 5]), sim=np.array([3, 3, 3, 3, 3])
    )
    b_dir = de.calc_bias_dir(brel_res)
    assert b_dir < 0


def test_bias_dir_for_equal_arrays():
    brel_res = de.calc_brel_res(obs=np.array([1, 2, 3]), sim=np.array([1, 2, 3]))
    b_dir = b_dir = de.calc_bias_dir(brel_res)
    assert b_dir == 0


def test_bias_slope_simulation_equals_obs_mean():
    brel_res = de.calc_brel_res(
        obs=np.array([1, 2, 3, 4, 5]), sim=np.array([3, 3, 3, 3, 3])
    )
    b_area = de.calc_bias_area(brel_res)
    b_dir = de.calc_bias_dir(brel_res)
    b_slope = de.calc_bias_slope(b_area, b_dir)
    assert b_slope > 0.5


def test_bias_slope_for_equal_arrays():
    brel_res = de.calc_brel_res(obs=np.array([1, 2, 3]), sim=np.array([1, 2, 3]))
    b_area = de.calc_bias_area(brel_res)
    b_dir = b_dir = de.calc_bias_dir(brel_res)
    b_slope = de.calc_bias_slope(b_area, b_dir)
    assert b_slope == 0


def test_arctan():
    phi = np.arctan2(0, 0)
    assert phi == 0


def test_fdc_sort():
    arr = np.array([1, 2, 3, 4, 5])
    arr_sort = np.sort(arr)
    assert arr_sort[0] < arr_sort[-1]


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_single, tolerance=DEFAULT_TOLERANCE
)
def test_single_diag_polar_plot():
    obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    fig = de.diag_polar_plot(obs, sim)
    return fig


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_multi, tolerance=DEFAULT_TOLERANCE
)
def test_multi_diag_polar_plot():
    brel_mean = np.array([0.1, 0.15, 0.2, 0.1, 0.05, 0.15])
    b_area = np.array([0.15, 0.1, 0.2, 0.1, 0.1, 0.2])
    temp_cor = np.array([0.9, 0.85, 0.8, 0.9, 0.85, 0.9])
    eff_de = np.array([0.79, 0.76, 0.65, 0.82, 0.81, 0.73])
    b_dir = np.array([0.08, 0.05, 0.1, 0.05, 0.05, 0.1])
    phi = np.array([0.58, 0.98, 0.78, 0.78, 0.46, 0.64])
    fig = de.diag_polar_plot_multi(brel_mean, b_area, temp_cor, eff_de, b_dir, phi)
    return fig



# -*- coding: utf-8 -*-

"""
de.fdc
~~~~~~~~~~~
Flow duration curve
:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

import numpy as np

# RunTimeWarning will not be displayed (division by zeros or NaN values)
np.seterr(divide="ignore", invalid="ignore")
import pandas as pd
import scipy as sp
import matplotlib.pyplot as plt
import seaborn as sns

# controlling figure aesthetics
sns.set_style("ticks", {"xtick.major.size": 8, "ytick.major.size": 8})
sns.set_context("paper", font_scale=1.5)


_mmd = r"[mm $d^{-1}$]"
_m3s = r"[$m^{3}$ $s^{-1}$]"
_q_lab = _mmd
_sim_lab = "Simulated"


def fdc(ts):
    """
    Flow duration curve for a single hydrologic time series.

    Parameters
    ----------
    ts : series
        Containing a hydrologic time series

    Returns
    ----------
    fig : Figure
        Returns a single figure containing flow duration curve.

    Examples
    --------
    Provide arrays with equal length

    >>> from de import fdc
    >>> import pandas as pd
    >>> date_rng = pd.date_range(start='1/1/2018', periods=11)
    >>> arr = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    >>> ts = pd.Series(data=arr, index=date_rng)
    >>> fdc.fdc(ts)
    """
    data = ts.dropna()
    data = np.sort(data.values)  # sort values by ascending order
    ranks = sp.stats.rankdata(data, method="ordinal")  # rank data
    ranks = ranks[::-1]
    # calculate exceedence probability
    prob = [(ranks[i] / (len(data) + 1)) for i in range(len(data))]

    fig, ax = plt.subplots()
    ax.plot(prob, data, color="blue")
    ax.set(ylabel=_q_lab, xlabel="Exceedence probabilty [-]", yscale="log")
    fig.subplots_adjust(left=0.2)

    return fig


def fdc_obs_sim(obs, sim):
    """
    Plotting the flow duration curves of two hydrologic time series (e.g.
    observed streamflow and simulated streamflow).

    Parameters
    ----------
    obs : series
        observed time series
    sim : series
        simulated time series

    Returns
    ----------
    fig : Figure
        Returns a single figure containing two flow duration curves.

    Examples
    --------
    Provide arrays with equal length

    >>> from de import fdc
    >>> import pandas as pd
    >>> date_rng = pd.date_range(start='1/1/2018', periods=11)
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    >>> ts_obs = pd.Series(data=obs, index=date_rng)
    >>> sim = np.array([1.4, .9, 1, 0.95, 1.4, 2.1, 2.6, 3.6, 1.9, 1.4, 1.1])
    >>> ts_sim = pd.Series(data=sim, index=date_rng)
    >>> fdc.fdc_obs_sim(ts_obs, ts_sim)
    """
    obs_sim = pd.DataFrame(index=obs.index, columns=["obs", "sim"])
    obs_sim.loc[:, "obs"] = obs.values
    obs_sim.loc[:, "sim"] = sim.values
    obs = obs_sim.sort_values(by=["obs"], ascending=True)
    sim = obs_sim.sort_values(by=["sim"], ascending=True)

    # calculate exceedence probability
    ranks_obs = sp.stats.rankdata(obs["obs"], method="ordinal")
    ranks_obs = ranks_obs[::-1]
    prob_obs = [(ranks_obs[i] / (len(obs["obs"]) + 1)) for i in range(len(obs["obs"]))]

    ranks_sim = sp.stats.rankdata(sim["sim"], method="ordinal")
    ranks_sim = ranks_sim[::-1]
    prob_sim = [(ranks_sim[i] / (len(sim["sim"]) + 1)) for i in range(len(sim["sim"]))]

    fig, ax = plt.subplots()
    ax.plot(prob_obs, obs["obs"], color="blue", lw=2, label="Observed")
    ax.plot(prob_sim, sim["sim"], color="red", lw=1, ls="-.", label=_sim_lab, alpha=0.8)
    ax.set(ylabel=_q_lab, xlabel="Exceedence probabilty [-]", yscale="log")
    ax.legend(loc=1)
    ax.set_xlim(0, 1)

    fig.subplots_adjust(left=0.2)

    return fig



# -*- coding: utf-8 -*-

"""
de.nse
~~~~~~~~~~~
Nash-Sutcliffe efficiency measure.
:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

from scipy import stats
import numpy as np

# RunTimeWarning will not be displayed (division by zeros or NaN values)
np.seterr(divide="ignore", invalid="ignore")


def calc_nse(obs, sim):
    r"""
    Calculate Nash-Sutcliffe-Efficiency (NSE).

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series as 1-D array

    Returns
    ----------
    sig : float
        Nash-Sutcliffe-Efficiency measure

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> nse.calc_nse(obs, sim)
    0.5648252536640361

    Notes
    ----------
    .. math::

        NSE = 1 - \frac{\sum_{t=1}^{t=T} (Q_{sim}(t) - Q_{obs}(t))^2}{\sum_{t=1}^{t=T} (Q_{obs}(t) - \overline{Q_{obs}})^2}


    References
    ----------
    Nash, J. E., and Sutcliffe, J. V.: River flow forecasting through
    conceptual models part I - A discussion of principles, Journal of
    Hydrology, 10, 282-290, 10.1016/0022-1694(70)90255-6, 1970.
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")
    sim_obs_diff = np.sum((sim - obs) ** 2)
    obs_mean = np.mean(obs)
    obs_diff_mean = np.sum((obs - obs_mean) ** 2)
    sig = 1 - (sim_obs_diff / obs_diff_mean)

    return sig


def calc_nse_dec(obs, sim):
    r"""
    Calculate the decomposed Nash-Sutcliffe-Efficiency (NSE).

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series as 1-D array

    Returns
    ----------
    sig : float
        decomposed Nash-Sutcliffe-Efficiency measure

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> nse.calc_nse_dec(obs, sim)
    0.9251076704923259

    Notes
    ----------
    .. math::

        NSE = 2 \times \alpha \times r - \alpha^2 - \beta_n^2


    References
    ----------
    Gupta, H. V., Kling, H., Yilmaz, K. K., and Martinez, G. F.: Decomposition
    of the mean squared error and NSE performance criteria: Implications for
    improving hydrological modelling, Journal of Hydrology, 377, 80-91,
    10.1016/j.jhydrol.2009.08.003, 2009.
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")
    nse_alpha = calc_nse_alpha(obs, sim)
    nse_beta = calc_nse_beta(obs, sim)
    nse_r = calc_nse_r(obs, sim)
    sig = 2 * nse_alpha * nse_r - nse_alpha ** 1 - nse_beta ** 2

    return sig


def calc_nse_beta(obs, sim):
    r"""
    Calculate the beta term of decomposed Nash-Sutcliffe-Efficiency (NSE).

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series as 1-D array

    Returns
    ----------
    nse_beta : float
        beta term of decomposed Nash-Sutcliffe-Efficiency (NSE)

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> nse.calc_nse_beta(obs, sim)
    0.2907828720750609

    Notes
    ----------
    .. math::

        \beta_{n} = \frac{\mu_{sim} - \mu_{obs}}{\sigma_{obs}}


    References
    ----------
    Gupta, H. V., Kling, H., Yilmaz, K. K., and Martinez, G. F.: Decomposition
    of the mean squared error and NSE performance criteria: Implications for
    improving hydrological modelling, Journal of Hydrology, 377, 80-91,
    10.1016/j.jhydrol.2009.08.003, 2009.
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")
    sim_mean = np.mean(sim)
    obs_mean = np.mean(obs)
    obs_std = np.std(obs)
    nse_beta = (sim_mean - obs_mean) / obs_std

    return nse_beta


def calc_nse_alpha(obs, sim):
    r"""
    Calculate the alpha term of decomposed Nash-Sutcliffe-Efficiency (NSE).

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series as 1-D array

    Returns
    ----------
    nse_alpha : float
        alpha term of decomposed Nash-Sutcliffe-Efficiency

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> nse.calc_nse_alpha(obs, sim)
    1.2812057455166919

    Notes
    ----------
    .. math::

        \alpha = \frac{\sigma_{sim}}{\sigma_{obs}}

    References
    ----------
    Gupta, H. V., Kling, H., Yilmaz, K. K., and Martinez, G. F.: Decomposition
    of the mean squared error and NSE performance criteria: Implications for
    improving hydrological modelling, Journal of Hydrology, 377, 80-91,
    10.1016/j.jhydrol.2009.08.003, 2009.
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")

    # calculate alpha term
    obs_std = np.std(obs)
    sim_std = np.std(sim)
    nse_alpha = sim_std / obs_std

    return nse_alpha


def calc_nse_r(obs, sim):
    """
    Calculate linear correlation between observed and simulated time series.

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series

    Returns
    ----------
    lin_cor : float
        Linear correlation between observed and simulated time series

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> nse.calc_nse_r(obs, sim)
    0.8940281850583509

    References
    ----------
    Gupta, H. V., Kling, H., Yilmaz, K. K., and Martinez, G. F.: Decomposition
    of the mean squared error and NSE performance criteria: Implications for
    improving hydrological modelling, Journal of Hydrology, 377, 80-91,
    10.1016/j.jhydrol.2009.08.003, 2009.
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")

    lin_cor = stats.pearsonr(obs, sim)[0]

    if np.isnan(lin_cor):
        lin_cor = 0

    return lin_cor



# -*- coding: utf-8 -*-

"""
de.signatures
~~~~~~~~~~~
Alternative signatures which can be used for a more specific diagnosis
(e.g. low flow behaviour)
:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

import numpy as np
# RunTimeWarning will not be displayed (division by zeros or NaN values)
np.seterr(divide='ignore', invalid='ignore')


def bias_rr(qobs, qsim):
    """
    Bias in runoff ratio.

    Parameters
    ----------
    qobs : (N,)array_like
        observed runoff time series as 1-D array

    qsim : (N,)array_like
        simulated runoff time series as 1-D array

    Returns
    ----------
    sig : float
        bias in runoff ratio (-)

    References
    ----------
    Yilmaz, K. K., Gupta, H. V., and Wagener, T.: A process-based diagnostic
    approach to model evaluation: Application to the NWS distributed hydrologic
    model, Water Resources Research, 44, 10.1029/2007wr006716, 2008.
    """
    if len(qobs) != len(qsim):
        raise AssertionError("Arrays are not of equal length!")

    sim_obs_diff = np.subtract(qsim, qobs)
    sum_diff = np.sum(sim_obs_diff)
    sum_obs = np.sum(qobs)
    sig = (sum_diff/sum_obs)

    return sig

def bias_hf(qobs, qsim):
    """
    Bias in high-flow segment of flow duration curve.

    Parameters
    ----------
    qobs : (N,)array_like
        observed runoff time series as 1-D array

    qsim : (N,)array_like
        simulated runoff time series as 1-D array

    Returns
    ----------
    sig : float
        bias in high-flow segment of flow duration curve (-)

    References
    ----------
    Yilmaz, K. K., Gupta, H. V., and Wagener, T.: A process-based diagnostic
    approach to model evaluation: Application to the NWS distributed hydrologic
    model, Water Resources Research, 44, 10.1029/2007wr006716, 2008.
    """
    if len(qobs) != len(qsim):
        raise AssertionError("Arrays are not of equal length!")

    qobs = np.sort(qobs)[::-1]
    qsim = np.sort(qsim)[::-1]

    hf_ind = int(.02 * len(qobs))
    qobs_hf = qobs[0:hf_ind]
    qsim_hf = qsim[0:hf_ind]

    diff_hf = np.subtract(qsim_hf, qobs_hf)
    sum_diff_hf = np.sum(diff_hf)
    sum_obs_hf = np.sum(qobs_hf)
    sig = (sum_diff_hf/sum_obs_hf)

    return sig

def bias_mf(qobs, qsim):
    """
    Bias in mid-flow segment of flow duration curve.

    Parameters
    ----------
    qobs : (N,)array_like
        observed runoff time series as 1-D array

    qsim : (N,)array_like
        simulated runoff time series as 1-D array

    Returns
    ----------
    sig : float
        bias in mid-flow segment of flow duration curve (-)

    References
    ----------
    Yilmaz, K. K., Gupta, H. V., and Wagener, T.: A process-based diagnostic
    approach to model evaluation: Application to the NWS distributed hydrologic
    model, Water Resources Research, 44, 10.1029/2007wr006716, 2008.
    """
    if len(qobs) != len(qsim):
        raise AssertionError("Arrays are not of equal length!")

    qobs = np.sort(qobs)[::-1]
    qsim = np.sort(qsim)[::-1]

    qobs_mf = np.log(np.median(qobs))
    qsim_mf = np.log(np.median(qsim))
    sig = ((qsim_mf - qobs_mf)/qobs_mf)

    return sig

def bias_lf(qobs, qsim):
    """
    Bias in low-flow segment of flow duration curve.

    Parameters
    ----------
    qobs : (N,)array_like
        observed runoff time series as 1-D array

    qsim : (N,)array_like
        simulated runoff time series as 1-D array

    Returns
    ----------
    sig : float
        bias in low-flow segment of flow duration curve (-)

    References
    ----------
    Yilmaz, K. K., Gupta, H. V., and Wagener, T.: A process-based diagnostic
    approach to model evaluation: Application to the NWS distributed hydrologic
    model, Water Resources Research, 44, 10.1029/2007wr006716, 2008.
    """
    if len(qobs) != len(qsim):
        raise AssertionError("Arrays are not of equal length!")

    qobs = np.sort(qobs)[::-1]
    qsim = np.sort(qsim)[::-1]

    l = int(.7 * len(qobs))
    L = len(qobs)
    qobs_lf = qobs[l:L]
    qsim_lf = qsim[l:L]

    diff_obs = np.log(qobs_lf) - np.log(qobs[L])
    diff_sim = np.log(qsim_lf) - np.log(qsim[L])

    sum_diff_obs = np.sum(diff_obs)
    sum_diff_sim = np.sum(diff_sim)
    sig = -1 * ((sum_diff_sim - sum_diff_obs)/sum_diff_obs)

    return sig

def bias_t(qobs, qsim, prec):
    """
    Bias in lag time.

    Parameters
    ----------
    qobs : (N,)array_like
        observed runoff time series as 1-D array

    qsim : (N,)array_like
        simulated runoff time series as 1-D array

    prec : (N,)array_like
        precipitation time series as 1-D array

    Returns
    ----------
    sig : float
        bias in lag time (-)

    References
    ----------
    Yilmaz, K. K., Gupta, H. V., and Wagener, T.: A process-based diagnostic
    approach to model evaluation: Application to the NWS distributed hydrologic
    model, Water Resources Research, 44, 10.1029/2007wr006716, 2008.
    """
    pass



# -*- coding: utf-8 -*-

"""
de.util
~~~~~~~~~~~

:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

import pandas as pd
import numpy as np
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import scipy as sp
import seaborn as sns
from de import de

# controlling figure aesthetics
sns.set_style("ticks", {"xtick.major.size": 8, "ytick.major.size": 8})


_mmd = r"[mm $d^{-1}$]"
_m3s = r"[$m^{3}$ $s^{-1}$]"
_q_lab = _mmd
_sim_lab = "Simulated"


def import_ts(path, sep=","):
    """
    Import .csv-file with streamflow time series (m3/s).

    Parameters
    ----------
    path : str
        Path to .csv-file which contains time series

    sep : str, optional
        Delimeter to use. The default is ‘,’.

    Returns
    ----------
    df_ts : dataframe
        Imported time series
    """
    df_ts = pd.read_csv(
        path, sep=sep, na_values=-9999, parse_dates=True, index_col=0, dayfirst=True
    )
    # drop nan values
    df_ts = df_ts.dropna()

    return df_ts


def import_camels_ts(path, sep=r"\s+", catch_area=None):
    r"""
    Import .txt-file with streamflow time series from CAMELS dataset (cubic
    feet per second).

    Parameters
    ----------
    path : str
        Path to .csv-file which contains time series

    sep : str, optional
        Delimeter to use. The default is r"\s+".

    catch_area : float, optional
        catchment area in km2 to convert runoff to mm/day

    Returns
    ----------
    df_ts : dataframe
        Imported time series in m3/s
    """
    if catch_area is None:
        raise ValueError("Value for catchment area is missing")
    df_ts = pd.read_csv(
        path, sep=sep, na_values=-999, header=None, parse_dates=[[1, 2, 3]], index_col=0
    )
    df_ts.drop(df_ts.columns[[0, 2]], axis=1, inplace=True)
    df_ts.columns = ["Qobs"]
    df_ts = df_ts.dropna()

    # convert to m3/s
    df_ts["Qobs"] = df_ts["Qobs"].values / 35.3
    # convert to mm/day
    df_ts["Qobs"] = (df_ts["Qobs"].values * (24 * 60 * 60) * 1000) / (
        catch_area * 1000 * 1000
    )

    return df_ts


def import_camels_obs_sim(path, sep=r"\s+"):
    r"""
    Import .txt-file containing observed and simulated streamflow time series
    from CAMELS dataset (mm per day).

    Parameters
    ----------
    path : str
        Path to .csv-file which contains time series

    sep : str, optional
        Delimeter to use. The default is r"\s+".

    Returns
    ----------
    obs_sim : dataframe
        observed and simulated time series in mm/day
    """
    obs_sim = pd.read_csv(
        path, sep=sep, na_values=-999, header=0, parse_dates=[[0, 1, 2]], index_col=0
    )
    obs_sim.drop(
        ["HR", "SWE", "PRCP", "RAIM", "TAIR", "PET", "ET"], axis=1, inplace=True
    )
    obs_sim.columns = ["Qsim", "Qobs"]
    obs_sim = obs_sim[["Qobs", "Qsim"]]
    obs_sim = obs_sim.dropna()

    return obs_sim


def plot_ts(ts):
    """Plot time series.

    Parameters
    ----------
    ts : dataframe
        Dataframe which contains time series

    Returns
    ----------
    fig : Figure
        time series plot

    >>> from de import util
    >>> import pandas as pd
    >>> date_rng = pd.date_range(start='1/1/2018', periods=11)
    >>> arr = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    >>> ts = pd.Series(data=arr, index=date_rng)
    >>> util.plot_ts(ts)
    """
    fig, ax = plt.subplots()
    ax.plot(ts.index, ts.values, color="blue")
    ax.set(ylabel=_q_lab, xlabel="Time [Years]")
    ax.set_ylim(0,)
    ax.set_xlim(ts.index[0], ts.index[-1])
    years_5 = mdates.YearLocator(5)
    years = mdates.YearLocator()
    yearsFmt = mdates.DateFormatter("%Y")
    ax.xaxis.set_major_locator(years_5)
    ax.xaxis.set_major_formatter(yearsFmt)
    ax.xaxis.set_minor_locator(years)

    return fig


def plot_obs_sim(obs, sim):
    """Plot observed and simulated time series.

    Parameters
    ----------
    obs : series
        observed time series

    sim : series
        simulated time series

    Returns
    ----------
    fig : Figure
        time series plot

    Examples
    --------
    Provide arrays with equal length

    >>> from de import util
    >>> import pandas as pd
    >>> date_rng = pd.date_range(start='1/1/2018', periods=11)
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    >>> ts_obs = pd.Series(data=obs, index=date_rng)
    >>> sim = np.array([1.4, .9, 1, 0.95, 1.4, 2.1, 2.6, 3.6, 1.9, 1.4, 1.1])
    >>> ts_sim = pd.Series(data=sim, index=date_rng)
    >>> util.plot_obs_sim(ts_obs, ts_sim)
    """
    fig, ax = plt.subplots()
    # observed time series
    ax.plot(obs.index, obs, lw=2, color="blue", label="Observed", alpha=0.8)
    # simulated time series
    ax.plot(sim.index, sim, lw=1, ls="-.", color="red", label=_sim_lab, alpha=0.9)
    ax.set(ylabel=_q_lab, xlabel="Time")
    ax.set_ylim(0,)
    ax.set_xlim(obs.index[0], obs.index[-1])

    # format the ticks
    years_20 = mdates.YearLocator(20)
    years_5 = mdates.YearLocator(5)
    yearsFmt = mdates.DateFormatter("%Y")
    ax.xaxis.set_major_locator(years_20)
    ax.xaxis.set_major_formatter(yearsFmt)
    ax.xaxis.set_minor_locator(years_5)

    return fig


def fdc_obs_sim_ax(obs, sim, ax, fig_num):  # pragma: no cover
    """Plotting the flow duration curves of two hydrologic time series (e.g.
    observed streamflow and simulated streamflow).

    Parameters
    ----------
    obs : series
        observed time series
    sim : series
        simulated time series
    ax : axes
        Axes object to draw the plot onto
    fig_num : string
        string object for figure caption
    """
    obs_sim = pd.DataFrame(index=obs.index, columns=["obs", "sim"])
    obs_sim.loc[:, "obs"] = obs.values
    obs_sim.loc[:, "sim"] = sim.values
    obs = obs_sim.sort_values(by=["obs"], ascending=True)
    sim = obs_sim.sort_values(by=["sim"], ascending=True)

    # calculate exceedence probability
    ranks_obs = sp.stats.rankdata(obs["obs"], method="ordinal")
    ranks_obs = ranks_obs[::-1]
    prob_obs = [(ranks_obs[i] / (len(obs["obs"]) + 1)) for i in range(len(obs["obs"]))]

    ranks_sim = sp.stats.rankdata(sim["sim"], method="ordinal")
    ranks_sim = ranks_sim[::-1]
    prob_sim = [(ranks_sim[i] / (len(sim["sim"]) + 1)) for i in range(len(sim["sim"]))]

    ax.plot(prob_obs, obs["obs"], lw=2, color="blue", alpha=0.5, label="Observed")
    ax.plot(prob_sim, sim["sim"], lw=2, ls="-.", color="red", label="Manipulated")
    ax.text(0.96, 0.95, fig_num, transform=ax.transAxes, ha="right", va="top")
    ax.set(yscale="log")
    ax.set_xlim(0, 1)


def plot_obs_sim_ax(obs, sim, ax, fig_num):  # pragma: no cover
    """Plot observed and simulated time series.

    Parameters
    ----------
    obs : series
        observed time series
    sim : series
        simulated time series
    ax : axes
        Axes object to draw the plot onto
    fig_num : string
        string object for figure caption
    """
    # observed time series
    ax.plot(obs.index, obs, lw=2, color="blue", label="Observed", alpha=0.8)
    ax.plot(
        sim.index, sim, lw=1.5, ls="-.", color="red", alpha=0.9, label=_sim_lab
    )  # simulated time series
    ax.set_ylim(0,)
    ax.set_xlim(obs.index[0], obs.index[-1])
    ax.text(0.96, 0.95, fig_num, transform=ax.transAxes, ha="right", va="top")
    # format the ticks
    years_20 = mdates.YearLocator(20)
    years_5 = mdates.YearLocator(5)
    yearsFmt = mdates.DateFormatter("%Y")
    ax.xaxis.set_major_locator(years_20)
    ax.xaxis.set_major_formatter(yearsFmt)
    ax.xaxis.set_minor_locator(years_5)


def diag_polar_plot_multi_fc(
    brel_mean, b_area, temp_cor, sig_de, b_dir, diag, fc, l=0.05, ax_lim=-0.6
):  # pragma: no cover
    r"""Multiple polar plot of Diagnostic-Efficiency (DE)

    Parameters
    ----------
    brel_mean : (N,)array_like
        relative mean bias as 1-D array

    b_area : (N,)array_like
        bias area as 1-D array

    temp_cor : (N,)array_like
        temporal correlation as 1-D array

    sig_de : (N,)array_like
        diagnostic efficiency as 1-D array

    b_dir : (N,)array_like
        direction of bias as 1-D array

    diag : (N,)array_like
        angle as 1-D array

    fc : list
        figure captions

    l : float, optional
        Threshold for which diagnosis can be made. The default is 0.05.

    Returns
    ----------
    fig : Figure
        diagnostic polar plot

    Notes
    ----------
    .. math::

        \varphi = arctan2(\overline{B_{rel}}, B_{slope})
    """
    sig_min = np.min(sig_de)

    ll_brel_mean = brel_mean.tolist()
    ll_b_dir = b_dir.tolist()
    ll_b_area = b_area.tolist()
    ll_sig = sig_de.tolist()
    ll_diag = diag.tolist()
    ll_temp_cor = temp_cor.tolist()

    # convert temporal correlation to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)

    delta = 0.01  # for spacing

    # determine axis limits
    yy = np.arange(ax_lim, 1.01, delta)[::-1]
    c_levels = np.arange(ax_lim, 1, 0.2)

    len_yy = 360
    # len_yy1 = 90

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    fig, ax = plt.subplots(
        figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
    )
    # dummie plot for colorbar of temporal correlation
    cs = np.arange(0, 1.1, 0.1)
    dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
    # Clear axis
    ax.cla()
    # plot regions
    ax.plot(
        (1, np.deg2rad(45)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(135)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(225)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(315)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    # contours of DE
    cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
    cl = ax.clabel(cp, inline=False, fontsize=10, fmt="%1.1f", colors="dimgrey")
    # threshold efficiency for FBM
    sig_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
    # loop over each data point
    for (bm, bd, ba, r, sig, ang, txt) in zip(
        ll_brel_mean, ll_b_dir, ll_b_area, ll_temp_cor, ll_sig, ll_diag, fc
    ):
        # slope of bias
        b_slope = de.calc_bias_slope(ba, bd)
        # convert temporal correlation to color
        rgba_color = cm.plasma_r(norm(r))
        # relation of b_dir which explains the error
        if abs(ba) > 0:
            exp_err = (abs(bd) * 2) / abs(ba)
        elif abs(ba) == 0:
            exp_err = 0

        # diagnose the error
        if abs(bm) <= l and exp_err > l and sig <= sig_l:
            c = ax.scatter(ang, sig, s=75, color=rgba_color, zorder=3)
            d = ax.scatter(ang, sig, color="grey", marker=".", zorder=4)
            ax.annotate(
                txt,
                xy=(ang, sig),
                color="black",
                fontsize=13,
                xytext=(-8, 0),
                textcoords="offset points",
                ha="center",
                va="center",
            )
        elif abs(bm) > l and exp_err <= l and sig <= sig_l:
            c = ax.scatter(ang, sig, s=75, color=rgba_color, zorder=3)
            d = ax.scatter(ang, sig, color="grey", marker=".", zorder=4)
            ax.annotate(
                txt,
                xy=(ang, sig),
                color="black",
                fontsize=13,
                xytext=(-8, 0),
                textcoords="offset points",
                ha="center",
                va="center",
            )
        elif abs(bm) > l and exp_err > l and sig <= sig_l:
            c = ax.scatter(ang, sig, s=75, color=rgba_color, zorder=3)
            d = ax.scatter(ang, sig, color="grey", marker=".", zorder=4)
            ax.annotate(
                txt,
                xy=(ang, sig),
                color="black",
                fontsize=13,
                xytext=(-8, 0),
                textcoords="offset points",
                ha="center",
                va="center",
            )
        # FBM
        elif abs(bm) <= l and exp_err <= l and sig <= sig_l:
            ax.annotate(
                "",
                xytext=(0, 1),
                xy=(0, sig),
                arrowprops=dict(edgecolor=rgba_color, facecolor="black", lw=3),
                zorder=2,
            )
            ax.annotate(
                "",
                xytext=(0, 1),
                xy=(np.pi, sig),
                arrowprops=dict(edgecolor=rgba_color, facecolor="black", lw=3),
                zorder=2,
            )
            ax.annotate(
                txt,
                xy=(ang, sig),
                color="black",
                fontsize=13,
                xytext=(-8, 0),
                textcoords="offset points",
                ha="center",
                va="center",
            )
        # FGM
        elif abs(bm) <= l and exp_err <= l and sig > sig_l:
            c = ax.scatter(ang, sig, s=75, color=rgba_color, zorder=3)
            d = ax.scatter(ang, sig, color="grey", marker=".", zorder=4)
            ax.annotate(
                txt,
                xy=(ang, sig),
                color="black",
                fontsize=13,
                xytext=(-6, 0),
                textcoords="offset points",
                ha="center",
                va="center",
            )
    ax.set_rticks([])  # turn default ticks off
    ax.set_rmin(1)
    ax.set_rmax(ax_lim)
    ax.tick_params(
        labelleft=False,
        labelright=False,
        labeltop=False,
        labelbottom=True,
        grid_alpha=0.01,
    )  # turn labels and grid off
    ax.set_xticklabels(["", "", "", "", "", "", "", ""])
    ax.text(
        -0.14,
        0.5,
        "High flow overestimation -",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        -0.09,
        0.5,
        "Low flow underestimation",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        -0.04,
        0.5,
        r"$B_{slope}$ < 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.14,
        0.5,
        "High flow underestimation -",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.09,
        0.5,
        "Low flow overestimation",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.04,
        0.5,
        r"$B_{slope}$ > 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        -0.09,
        "Constant negative offset",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        -0.04,
        r"$\overline{B_{rel}}$ < 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        1.09,
        "Constant positive offset",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        1.04,
        r"$\overline{B_{rel}}$ > 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    # add colorbar for temporal correlation
    cbar = fig.colorbar(
        dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
    )
    cbar.set_label(r"r [-]", fontsize=12, labelpad=8)
    cbar.set_ticklabels(["1", "0.5", "<0"])
    cbar.ax.tick_params(direction="in", labelsize=10)

    return fig


def polar_plot_multi_fc(
    kge_beta, alpha_or_gamma, kge_r, sig_kge, fc, ax_lim=-0.6
):  # pragma: no cover
    """Polar plot of Kling-Gupta efficiency (KGE) with multiple
    values.

    Parameters
    ----------
    kge_beta: (N,)array_like
        KGE beta as 1-D array

    alpha_or_gamma : (N,)array_like
        KGE alpha or KGE gamma as 1-D array

    kge_r : (N,)array_like
        KGE r as 1-D array

    sig_kge : (N,)array_like
        KGE as 1-D array

    fc : list
        figure captions

    Returns
    ----------
    fig : Figure
        diagnostic polar plot
    """
    sig_min = np.min(sig_kge)

    ll_kge_beta = kge_beta.tolist()
    ll_ag = alpha_or_gamma.tolist()
    ll_kge_r = kge_r.tolist()
    ll_sig = sig_kge.tolist()

    # convert temporal correlation to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)

    delta = 0.01  # for spacing

    # determine axis limits
    yy = np.arange(ax_lim, 1.01, delta)[::-1]
    c_levels = np.arange(ax_lim, 1, 0.2)

    len_yy = 360
    # len_yy1 = 90

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    fig, ax = plt.subplots(
        figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
    )
    # dummie plot for colorbar of temporal correlation
    cs = np.arange(0, 1.1, 0.1)
    dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
    # Clear axis
    ax.cla()
    # plot regions
    ax.plot(
        (1, np.deg2rad(45)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(135)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(225)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(315)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    # contours of KGE
    cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
    cl = ax.clabel(cp, inline=False, fontsize=10, fmt="%1.1f", colors="dimgrey")
    # loop over each data point
    for (b, ag, r, sig, txt) in zip(ll_kge_beta, ll_ag, ll_kge_r, ll_sig, fc):
        ang = np.arctan2(b - 1, ag - 1)
        # convert temporal correlation to color
        rgba_color = cm.plasma_r(norm(r))
        c = ax.scatter(ang, sig, s=75, color=rgba_color, zorder=2)
        d = ax.scatter(ang, sig, color="grey", marker=".", zorder=4)
        ax.annotate(
            txt,
            xy=(ang, sig),
            color="black",
            fontsize=13,
            xytext=(8, 0),
            textcoords="offset points",
            ha="center",
            va="center",
        )
    ax.tick_params(
        labelleft=False,
        labelright=False,
        labeltop=False,
        labelbottom=True,
        grid_alpha=0.01,
    )  # turn labels and grid off
    ax.text(
        -0.09,
        0.5,
        "Variability underestimation",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        -0.04,
        0.5,
        r"($\alpha$ - 1) < 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.09,
        0.5,
        "Variability overestimation",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.04,
        0.5,
        r"($\alpha$ - 1) > 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        -0.09,
        "Mean underestimation",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        -0.04,
        r"($\beta$ - 1) < 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        1.09,
        "Mean overestimation",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        1.04,
        r"($\beta$ - 1) > 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.set_xticklabels(["", "", "", "", "", "", "", ""])
    ax.set_rticks([])  # turn default ticks off
    ax.set_rmin(1)
    ax.set_rmax(ax_lim)
    # add colorbar for temporal correlation
    cbar = fig.colorbar(
        dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
    )
    cbar.set_label(r"r [-]", fontsize=12, labelpad=8)
    cbar.set_ticklabels(["1", "0.5", "<0"])
    cbar.ax.tick_params(direction="in", labelsize=10)

    return fig



# -*- coding: utf-8 -*-

"""
de.de
~~~~~~~~~~~
Diagnosing model performance using an efficiency measure based on flow
duration curve and temporal correlation. The efficiency measure can be
visualized by diagnostic polar plots which facilitates decomposing potential
error contributions (dynamic errors vs. constant erros vs. timing errors)
:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

import numpy as np

# RunTimeWarning will not be displayed (division by zeros or NaN values)
np.seterr(divide="ignore", invalid="ignore")
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import pandas as pd
import scipy as sp
import scipy.integrate as integrate
import seaborn as sns

# controlling figure aesthetics
sns.set_style("ticks", {"xtick.major.size": 8, "ytick.major.size": 8})
sns.set_context("paper", font_scale=1.5)


def calc_brel_mean(obs, sim, sort=True):
    r"""
    Calculate arithmetic mean of relative bias.

    Parameters
    ----------
    obs : (N,)array_like
        observed time series as 1-D array

    sim : (N,)array_like
        simulated time series

    sort : boolean, optional
        If True, time series are sorted by ascending order. If False, time
        series are not sorted. The default is to sort.

    Returns
    ----------
    brel_mean : float
        average relative bias

    Notes
    ----------
    .. math::

        \overline{B_{rel}} = \frac{1}{N}\sum_{i=1}^{N} B_{rel}(i)


    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> de.calc_brel_mean(obs, sim)
    0.09330065359477124
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")

    if sort:
        obs = np.sort(obs)[::-1]
        sim = np.sort(sim)[::-1]

    sim_obs_diff = np.subtract(sim, obs)
    brel = np.divide(sim_obs_diff, obs)
    brel[sim_obs_diff == 0] = 0
    brel = brel[np.isfinite(brel)]
    brel_mean = np.mean(brel)

    return brel_mean


def calc_brel_res(obs, sim, sort=True):
    r"""
    Subtract arithmetic mean of relative bias from relative bias.

    Parameters
    ----------
    obs : (N,)array_like
        observed time series as 1-D array

    sim : (N,)array_like
        simulated time series

    sort : boolean, optional
        If True, time series are sorted by ascending order. If False, time
        series are not sorted. The default is to sort.

    Returns
    ----------
    brel_res : array_like
        remaining relative bias

    Notes
    ----------
    .. math::

        B_{rest}(i) = B_{rel}(i) - \overline{B_{rel}}

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> de.calc_brel_res(obs, sim)
    array([ 0.15669935, -0.02663399, -0.22663399,  0.10669935,  0.08316993,
       -0.09330065])
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")

    if sort:
        obs = np.sort(obs)[::-1]
        sim = np.sort(sim)[::-1]
    sim_obs_diff = np.subtract(sim, obs)
    brel = np.divide(sim_obs_diff, obs)
    brel[sim_obs_diff == 0] = 0
    brel = brel[np.isfinite(brel)]
    brel_mean = np.mean(brel)
    brel_res = brel - brel_mean

    return brel_res


def calc_bias_area(brel_res):
    r"""
    Calculate absolute bias area for entire flow duration curve.

    Parameters
    ----------
    brel_res : (N,)array_like
        remaining relative bias as 1-D array

    Returns
    ----------
    b_area : float
        bias area

    Notes
    ----------
    .. math::

        \vert B_{area}\vert = \int_{0}^{1}\vert B_{rest}(i)\vert di

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> b_res = de.calc_brel_res(obs, sim)
    >>> de.calc_bias_area(b_res)
    0.1112908496732026

    See Also
    --------
    de.calc_brel_res
    """
    perc = np.linspace(0, 1, len(brel_res))
    # area of absolute bias
    b_area = integrate.simps(abs(brel_res), perc)

    return b_area


def calc_bias_dir(brel_res):
    r"""
    Calculate absolute bias area for high flow and low flow.

    Parameters
    ----------
    brel_res : (N,)array_like
        remaining relative bias as 1-D array

    Returns
    ----------
    b_dir : float
        direction of bias

    Notes
    ----------
    .. math::

        B_{dir} = \int_{0}^{0.5}B_{rest}(i) di

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> b_res = de.calc_brel_res(obs, sim)
    >>> de.calc_bias_dir(b_res)
    -0.014705882352941155
    """
    mid_idx = int(len(brel_res) / 2)
    # integral of relative bias < 50 %
    perc = np.linspace(0, 0.5, mid_idx)
    # direction of bias
    b_dir = integrate.simps(brel_res[:mid_idx], perc)

    return b_dir


def calc_bias_slope(b_area, b_dir):
    r"""
    Calculate slope of bias balance.

    Parameters
    ----------
    b_area : float
        absolute area of remaining bias

    b_dir : float
        direction of bias

    Returns
    ----------
    b_slope : float
        slope of bias

    Notes
    ----------
    .. math::

        B_{slope} =
        \begin{cases}
        \vert B_{area}\vert \times (-1) & \text{if } B_{dir} > 0 \\
        \vert B_{area}\vert       & \text{if } B_{dir} < 0 \\
        0       & \text{if } B_{dir} = 0
        \end{cases}

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> b_res = de.calc_brel_res(obs, sim)
    >>> b_dir = de.calc_bias_dir(b_res)
    >>> b_area = de.calc_bias_area(b_res)
    >>> de.calc_bias_slope(b_area, b_dir)
    0.1112908496732026
    """
    if b_dir > 0:
        b_slope = b_area * (-1)

    elif b_dir < 0:
        b_slope = b_area

    elif b_dir == 0:
        b_slope = 0

    return b_slope


def calc_temp_cor(obs, sim, r="pearson"):
    """
    Calculate temporal correlation between observed and simulated
    time series.

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series

    r : str, optional
        Either Spearman correlation coefficient ('spearman') or Pearson
        correlation coefficient ('pearson') can be used to describe the temporal
        correlation. The default is to calculate the Pearson correlation.

    Returns
    ----------
    temp_cor : float
        correlation between observed and simulated time series

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> de.calc_temp_cor(obs, sim)
    0.8940281850583509
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")

    if r == "spearman":
        r = sp.stats.spearmanr(obs, sim)
        temp_cor = r[0]

        if np.isnan(temp_cor):
            temp_cor = 0

    elif r == "pearson":
        r = sp.stats.pearsonr(obs, sim)
        temp_cor = r[0]

        if np.isnan(temp_cor):
            temp_cor = 0

    return temp_cor


def calc_de(obs, sim, sort=True):
    r"""
    Calculate Diagnostic-Efficiency (DE).

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series

    sort : boolean, optional
        If True, time series are sorted by ascending order. If False, time
        series are not sorted. The default is to sort.

    Returns
    ----------
    eff : float
        non-normalized Diagnostic efficiency

    Notes
    ----------
    .. math::

        DE = 1 - \sqrt{\overline{B_{rel}}^2 + \vert B_{area}\vert^2 + (r - 1)^2}
    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> de.calc_de(obs, sim)
    0.8202204384691575
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")
    # mean relative bias
    brel_mean = calc_brel_mean(obs, sim, sort=sort)
    # remaining relative bias
    brel_res = calc_brel_res(obs, sim, sort=sort)
    # area of relative remaing bias
    b_area = calc_bias_area(brel_res)
    # temporal correlation
    temp_cor = calc_temp_cor(obs, sim)
    # diagnostic efficiency
    eff = 1 - np.sqrt((brel_mean) ** 2 + (b_area) ** 2 + (temp_cor - 1) ** 2)

    return eff


def diag_polar_plot(obs, sim, sort=True, l=0.05, extended=False):
    r"""
    Diagnostic polar plot of Diagnostic efficiency (DE) for a single
    evaluation.

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series as 1-D array

    sort : boolean, optional
        If True, time series are sorted by ascending order. If False, time
        series are not sorted. The default is to sort.

    l : float, optional
        Threshold for which diagnosis can be made. The default is 0.05.

    extended : boolean, optional
        If True, extended diagnostic plot is displayed. In addtion, the
        duration curve of B_rest is plotted besides the polar plot. The default
        is, that only the diagnostic polar plot is displayed.

    Returns
    ----------
    fig : Figure
        Returns a single figure if extended=False and two figures if
        extended=True.

    Notes
    ----------
    .. math::

        \varphi = arctan2(\overline{B_{rel}}, B_{slope})

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    >>> sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    >>> de.diag_polar_plot(obs, sim)
    """
    if len(obs) != len(sim):
        raise AssertionError("Arrays are not of equal length!")
    # mean relative bias
    brel_mean = calc_brel_mean(obs, sim, sort=sort)

    # remaining relative bias
    brel_res = calc_brel_res(obs, sim, sort=sort)
    # area of relative remaing bias
    b_area = calc_bias_area(brel_res)
    # temporal correlation
    temp_cor = calc_temp_cor(obs, sim)
    # diagnostic efficiency
    eff = 1 - np.sqrt((brel_mean) ** 2 + (b_area) ** 2 + (temp_cor - 1) ** 2)

    # direction of bias
    b_dir = calc_bias_dir(brel_res)

    # slope of bias
    b_slope = calc_bias_slope(b_area, b_dir)

    # convert to radians
    # (y, x) Trigonometric inverse tangent
    phi = np.arctan2(brel_mean, b_slope)

    # convert temporal correlation to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)
    rgba_color = cm.plasma_r(norm(temp_cor))

    delta = 0.01  # for spacing

    # determine axis limits
    if eff >= 0:
        ax_lim = 0.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(0, 1, 0.2)
    elif eff < 0 and eff >= -1:
        ax_lim = 1.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-1, 1, 0.2)
    elif eff >= -2 and eff < -1:
        ax_lim = 2.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-2, 1, 0.2)
    elif eff <= -2:
        raise AssertionError("Value of 'DE' is out of bounds for visualization!", eff)

    len_yy = len(yy)

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    if not extended:
        fig, ax = plt.subplots(
            figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
        )
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # relation of b_dir which explains the error
        if abs(b_area) > 0:
            exp_err = (abs(b_dir) * 2) / abs(b_area)
        elif abs(b_area) == 0:
            exp_err = 0
        # diagnose the error
        if abs(brel_mean) <= l and exp_err > l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        elif abs(brel_mean) > l and exp_err <= l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        elif abs(brel_mean) > l and exp_err > l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        # FBM
        elif abs(brel_mean) <= l and exp_err <= l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(0, eff), arrowprops=dict(facecolor=rgba_color)
            )
            ax.annotate(
                "",
                xytext=(0, 1),
                xy=(np.pi, eff),
                arrowprops=dict(facecolor=rgba_color),
            )
        # FGM
        elif abs(brel_mean) <= l and exp_err <= l and eff > eff_l:
            c = ax.scatter(phi, eff, color=rgba_color)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff >= 0:
            ax.set_rmax(0)
        elif eff < 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", "", "", ""])
        ax.text(
            -0.14,
            0.5,
            "High flow overestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.09,
            0.5,
            "Low flow underestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.04,
            0.5,
            r"$B_{slope}$ < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.14,
            0.5,
            "High flow underestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.09,
            0.5,
            "Low flow overestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"$B_{slope}$ > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.09,
            "Constant negative offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"$\overline{B_{rel}}$ < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.09,
            "Constant positive offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"$\overline{B_{rel}}$ > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("r [-]", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        return fig

    elif extended:
        fig = plt.figure(figsize=(12, 6), constrained_layout=True)
        gs = fig.add_gridspec(1, 2)
        ax = fig.add_subplot(gs[0, 0], projection="polar")
        ax1 = fig.add_axes([0.65, 0.3, 0.33, 0.33], frameon=True)
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # relation of b_dir which explains the error
        if abs(b_area) > 0:
            exp_err = (abs(b_dir) * 2) / abs(b_area)
        elif abs(b_area) == 0:
            exp_err = 0
        # diagnose the error
        if abs(brel_mean) <= l and exp_err > l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        elif abs(brel_mean) > l and exp_err <= l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        elif abs(brel_mean) > l and exp_err > l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
            )
        # FBM
        elif abs(brel_mean) <= l and exp_err <= l and eff <= eff_l:
            ax.annotate(
                "", xytext=(0, 1), xy=(0, eff), arrowprops=dict(facecolor=rgba_color)
            )
            ax.annotate(
                "",
                xytext=(0, 1),
                xy=(np.pi, eff),
                arrowprops=dict(facecolor=rgba_color),
            )
        # FGM
        elif abs(brel_mean) <= l and exp_err <= l and eff > eff_l:
            c = ax.scatter(phi, eff, color=rgba_color)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff >= 0:
            ax.set_rmax(0)
        elif eff < 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", "", "", ""])
        ax.text(
            -0.14,
            0.5,
            "High flow overestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.09,
            0.5,
            "Low flow underestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.04,
            0.5,
            r"$B_{slope}$ < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.14,
            0.5,
            "High flow underestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.09,
            0.5,
            "Low flow overestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"$B_{slope}$ > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.09,
            "Constant negative offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"$\overline{B_{rel}}$ < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.09,
            "Constant positive offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"$\overline{B_{rel}}$ > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("r [-]", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        # plot B_rest
        # calculate exceedence probability
        prob = np.linspace(0, 1, len(brel_res))
        ax1.axhline(y=0, color="slategrey")
        ax1.axvline(x=0.5, color="slategrey")
        ax1.plot(prob, brel_res, color="black")
        ax1.fill_between(prob, brel_res, where=0 < brel_res, facecolor="purple")
        ax1.fill_between(prob, brel_res, where=0 > brel_res, facecolor="red")
        ax1.set(ylabel=r"$B_{rest}$ [-]", xlabel="Exceedence probabilty [-]")

        return fig


def diag_polar_plot_multi(
    brel_mean, b_area, temp_cor, eff_de, b_dir, phi, l=0.05, extended=False
):
    r"""
    Diagnostic polar plot of Diagnostic efficiency (DE) for multiple
    evaluations.

    Note that points are used rather than arrows. Displaying multiple
    arrows would deteriorate visual comprehension.

    Parameters
    ----------
    brel_mean : (N,)array_like
        relative mean bias as 1-D array

    b_area : (N,)array_like
        bias area as 1-D array

    temp_cor : (N,)array_like
        temporal correlation as 1-D array

    eff_de : (N,)array_like
        diagnostic efficiency as 1-D array

    b_dir : (N,)array_like
        direction of bias as 1-D array

    phi : (N,)array_like
        angle as 1-D array

    l : float, optional
        Deviation of metric terms used to calculate the threshold of DE for
        which diagnosis is available. The default is 0.05.

    extended : boolean, optional
        If True, density plot is displayed. In addtion, the density plot
        is displayed besides the polar plot. The default is,
        that only the diagnostic polar plot is displayed.

    Returns
    ----------
    fig : Figure
        diagnostic polar plot

    Notes
    ----------
    .. math::

        \varphi = arctan2(\overline{B_{rel}}, B_{slope})

    Examples
    --------
    Provide arrays with equal length

    >>> from de import de
    >>> import numpy as np
    >>> brel_mean = np.array([0.1, 0.15, 0.2, 0.1, 0.05, 0.15])
    >>> b_area = np.array([0.15, 0.1, 0.2, 0.1, 0.1, 0.2])
    >>> temp_cor = np.array([0.9, 0.85, 0.8, 0.9, 0.85, 0.9])
    >>> eff_de = np.array([0.79, 0.76, 0.65, 0.82, 0.81, 0.73])
    >>> b_dir = np.array([0.08, 0.05, 0.1, 0.05, 0.05, 0.1])
    >>> phi = np.array([0.58, 0.98, 0.78, 0.78, 0.46, 0.64])
    >>> de.diag_polar_plot_multi(brel_mean, b_area, temp_cor, eff_de, b_dir,
                                 phi)
    """
    eff_min = np.min(eff_de)

    ll_brel_mean = brel_mean.tolist()
    ll_b_dir = b_dir.tolist()
    ll_b_area = b_area.tolist()
    ll_eff = eff_de.tolist()
    ll_phi = phi.tolist()
    ll_temp_cor = temp_cor.tolist()

    # convert temporal correlation to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)

    delta = 0.01  # for spacing

    # determine axis limits
    if eff_min >= 0:
        ax_lim = 0.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(0, 1, 0.2)
    elif eff_min < 0 and eff_min >= -1:
        ax_lim = 1.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-1, 1, 0.2)
    elif eff_min >= -2 and eff_min < -1:
        ax_lim = 2.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-2, 1, 0.2)
    elif eff_min <= -2:
        raise ValueError("Some values of 'DE' are too low for visualization!", eff_min)

    len_yy = len(yy)

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    if not extended:
        fig, ax = plt.subplots(
            figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
        )
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # loop over each data point
        zz = zip(ll_brel_mean, ll_b_dir, ll_b_area, ll_temp_cor, ll_eff, ll_phi)
        for (bm, bd, ba, r, eff, ang) in zz:
            # slope of bias
            b_slope = calc_bias_slope(ba, bd)
            # convert temporal correlation to color
            rgba_color = cm.plasma_r(norm(r))
            # relation of b_dir which explains the error
            if abs(ba) > 0:
                exp_err = (abs(bd) * 2) / abs(ba)
            elif abs(ba) == 0:
                exp_err = 0
            # diagnose the error
            if abs(bm) <= l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(bm) > l and exp_err <= l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(bm) > l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            # FBM
            elif abs(bm) <= l and exp_err <= l and eff <= eff_l:
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(0, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(np.pi, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
            # FGM
            elif abs(bm) <= l and exp_err <= l and eff > eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff_min >= 0:
            ax.set_rmax(0)
        elif eff_min <= 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", "", "", ""])
        ax.text(
            -0.14,
            0.5,
            "High flow overestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.09,
            0.5,
            "Low flow underestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.04,
            0.5,
            r"$B_{slope}$ < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.14,
            0.5,
            "High flow underestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.09,
            0.5,
            "Low flow overestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"$B_{slope}$ > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.09,
            "Constant negative offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"$\overline{B_{rel}}$ < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.09,
            "Constant positive offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"$\overline{B_{rel}}$ > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("r [-]", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        return fig

    elif extended:
        fig = plt.figure(figsize=(12, 6), constrained_layout=True)
        gs = fig.add_gridspec(1, 2)
        ax = fig.add_subplot(gs[0, 0], projection="polar")
        ax1 = fig.add_axes([0.66, 0.3, 0.32, 0.32], frameon=True)
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # loop over each data point
        zz = zip(ll_brel_mean, ll_b_dir, ll_b_area, ll_temp_cor, ll_eff, ll_phi)
        for (bm, bd, ba, r, eff, ang) in zz:
            # slope of bias
            b_slope = calc_bias_slope(ba, bd)
            # convert temporal correlation to color
            rgba_color = cm.plasma_r(norm(r))
            # relation of b_dir which explains the error
            if abs(ba) > 0:
                exp_err = (abs(bd) * 2) / abs(ba)
            elif abs(ba) == 0:
                exp_err = 0
            # diagnose the error
            if abs(bm) <= l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(bm) > l and exp_err <= l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(bm) > l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            # FBM
            elif abs(bm) <= l and exp_err <= l and eff <= eff_l:
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(0, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(np.pi, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
            # FGM
            elif abs(bm) <= l and exp_err <= l and eff > eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff_min >= 0:
            ax.set_rmax(0)
        elif eff_min <= 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", r"0$^{\circ}$ (360$^{\circ}$)", "", ""])
        ax.text(
            -0.14,
            0.5,
            "High flow overestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.09,
            0.5,
            "Low flow underestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            -0.04,
            0.5,
            r"$B_{slope}$ < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.14,
            0.5,
            "High flow underestimation -",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.09,
            0.5,
            "Low flow overestimation",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"$B_{slope}$ > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.09,
            "Constant negative offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"$\overline{B_{rel}}$ < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.09,
            "Constant positive offset",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"$\overline{B_{rel}}$ > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("r [-]", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        # convert to degrees
        diag_deg = (phi * (180 / np.pi)) + 135
        diag_deg[diag_deg < 0] = 360 - diag_deg[diag_deg < 0]

        # 1-D density plot
        g = sns.kdeplot(diag_deg, color="k", ax=ax1)
        ax1.set_xticks([0, 90, 180, 270, 360])
        ax1.set_xlim(0, 360)
        ax1.set_ylim(0,)
        ax1.set(ylabel="Density", xlabel=r"[$^\circ$]")

        # 2-D density plot
        r_colors = cm.plasma_r(norm(temp_cor))
        g = sns.jointplot(
            diag_deg,
            eff_de,
            kind="kde",
            zorder=1,
            n_levels=20,
            cmap="Greys",
            shade_lowest=False,
            marginal_kws={"color": "k", "shade": False},
        ).plot_joint(plt.scatter, c=r_colors, alpha=0.4, zorder=2)
        g.set_axis_labels(r"[$^\circ$]", "DE [-]")
        g.ax_joint.set_xticks([0, 90, 180, 270, 360])
        g.ax_joint.set_xlim(0, 360)
        g.ax_joint.set_ylim(-ax_lim, 1)
        g.ax_marg_x.set_xticks([0, 90, 180, 270, 360])
        kde_data = g.ax_marg_x.get_lines()[0].get_data()
        kde_xx = kde_data[0]
        kde_yy = kde_data[1]
        kde_data = g.ax_marg_y.get_lines()[0].get_data()
        kde_xx = kde_data[0]
        kde_yy = kde_data[1]
        norm = matplotlib.colors.Normalize(vmin=-ax_lim, vmax=1.0)
        colors = cm.Reds_r(norm(kde_yy))
        npts = len(kde_xx)
        for i in range(npts - 1):
            g.ax_marg_y.fill_betweenx(
                [kde_yy[i], kde_yy[i + 1]], [kde_xx[i], kde_xx[i + 1]], color=colors[i]
            )
        g.fig.tight_layout()

        return fig, g.fig


def gdiag_polar_plot(eff, comp1, comp2, comp3, l=0.05):  # pragma: no cover
    r"""
    Generic diagnostic polar plot for  single evaluation.

    Parameters
    ----------
    eff : float
        efficiency measure

    comp1 : float
        metric component 1

    comp2 : float
        metric component 2

    comp3 : float
        metric component 3

    l : float, optional
        Deviation of metric terms used to calculate the threshold of DE for
        which diagnosis is available. The default is 0.05.

    Returns
    ----------
    fig : Figure
        diagnostic polar plot

    Notes
    ----------
    .. math::

        \varphi = arctan2(comp1, comp2)
    """
    # convert to radians
    # (y, x) Trigonometric inverse tangent
    phi = np.arctan2(comp1, comp2)

    # convert metric component 3 to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)
    rgba_color = cm.plasma_r(norm(comp3))

    delta = 0.01  # for spacing

    # determine axis limits
    if eff >= 0:
        ax_lim = 0.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(0, 1, 0.2)
    elif eff < 0 and eff >= -1:
        ax_lim = 1.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-1, 1, 0.2)
    elif eff >= -2 and eff < -1:
        ax_lim = 2.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-2, 1, 0.2)
    elif eff <= -2:
        raise AssertionError("Value of eff is out of bounds for visualization!", eff)

    len_yy = len(yy)

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    fig, ax = plt.subplots(
        figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
    )
    # dummie plot for colorbar of temporal correlation
    cs = np.arange(0, 1.1, 0.1)
    dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
    # Clear axis
    ax.cla()
    # plot regions
    ax.plot(
        (1, np.deg2rad(45)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(135)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(225)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    ax.plot(
        (1, np.deg2rad(315)),
        (1, np.min(yy)),
        color="lightgray",
        linewidth=1.5,
        ls="--",
        zorder=0,
    )
    # contours of DE
    cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
    cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
    # threshold efficiency for FBM
    eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
    # relation of b_dir which explains the error
    if abs(comp2) > 0:
        exp_err = comp2
    elif abs(comp2) == 0:
        exp_err = 0
    # diagnose the error
    if abs(comp1) <= l and exp_err > l and eff <= eff_l:
        ax.annotate(
            "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
        )
    elif abs(comp1) > l and exp_err <= l and eff <= eff_l:
        ax.annotate(
            "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
        )
    elif abs(comp1) > l and exp_err > l and eff <= eff_l:
        ax.annotate(
            "", xytext=(0, 1), xy=(phi, eff), arrowprops=dict(facecolor=rgba_color)
        )
    # FBM
    elif abs(comp1) <= l and exp_err <= l and eff <= eff_l:
        ax.annotate(
            "", xytext=(0, 1), xy=(0, eff), arrowprops=dict(facecolor=rgba_color)
        )
        ax.annotate(
            "", xytext=(0, 1), xy=(np.pi, eff), arrowprops=dict(facecolor=rgba_color)
        )
    # FGM
    elif abs(comp1) <= l and exp_err <= l and eff > eff_l:
        c = ax.scatter(phi, eff, color=rgba_color)
    ax.set_rticks([])  # turn default ticks off
    ax.set_rmin(1)
    if eff >= 0:
        ax.set_rmax(0)
    elif eff <= 0:
        ax.set_rmax(-ax_lim + 0.2)
    # turn labels and grid off
    ax.tick_params(
        labelleft=False,
        labelright=False,
        labeltop=False,
        labelbottom=True,
        grid_alpha=0.01,
    )
    ax.set_xticklabels(["", "", "", "", "", "", "", ""])
    ax.text(
        -0.04,
        0.5,
        r"Comp2 < 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        1.04,
        0.5,
        r"Comp2 > 0",
        va="center",
        ha="center",
        rotation=90,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        -0.04,
        r"Comp1 < 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    ax.text(
        0.5,
        1.04,
        r"Comp1 > 0",
        va="center",
        ha="center",
        rotation=0,
        rotation_mode="anchor",
        transform=ax.transAxes,
    )
    # add colorbar for temporal correlation
    cbar = fig.colorbar(
        dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
    )
    cbar.set_label("Comp3", labelpad=4)
    cbar.set_ticklabels(["1", "0.5", "<0"])
    cbar.ax.tick_params(direction="in")

    return fig


def gdiag_polar_plot_multi(
    eff, comp1, comp2, comp3, l=0.05, extended=True
):  # pragma: no cover
    r"""
    Generic diagnostic polar plot for multiple evaluations.

    Parameters
    ----------
    eff : (N,)array_like
        efficiency measure

    comp1 : (N,)array_like
        metric component 1

    comp2 : (N,)array_like
        metric component 2

    comp3 : (N,)array_like
        metric component 3

    l : float, optional
        Deviation of metric terms used to calculate the threshold of DE for
        which diagnosis is available. The default is 0.05.

    extended : boolean, optional
        If True, density plot is displayed. In addtion, the density plot
        is displayed besides the polar plot. The default is,
        that only the diagnostic polar plot is displayed.

    Returns
    ----------
    fig : Figure
        diagnostic polar plot

    Notes
    ----------
    .. math::

        \varphi = arctan2(comp1, comp2)
    """
    eff_min = np.min(eff)

    ll_comp1 = comp1.tolist()
    ll_comp2 = comp2.tolist()
    ll_comp3 = comp3.tolist()
    ll_eff = eff.tolist()

    # convert temporal correlation to color
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)

    delta = 0.01  # for spacing

    # determine axis limits
    if eff_min >= 0:
        ax_lim = 0.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(0, 1, 0.2)
    elif eff_min < 0 and eff_min >= -1:
        ax_lim = 1.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-1, 1, 0.2)
    elif eff_min >= -2 and eff_min < -1:
        ax_lim = 2.2
        yy = np.arange(-ax_lim, 1.01, delta)[::-1]
        c_levels = np.arange(-2, 1, 0.2)
    elif eff_min <= -2:
        raise ValueError(
            "Some values of eff are out of bounds for visualization!", eff_min
        )

    len_yy = len(yy)

    # arrays to plot contour lines of DE
    xx = np.radians(np.linspace(0, 360, len_yy))
    theta, r = np.meshgrid(xx, yy)

    # diagnostic polar plot
    if not extended:
        fig, ax = plt.subplots(
            figsize=(6, 6), subplot_kw=dict(projection="polar"), constrained_layout=True
        )
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # loop over each data point
        for (c1, c2, c3, eff) in zip(ll_comp1, ll_comp2, ll_comp3, ll_eff):
            ang = np.arctan2(c1, c2)
            # convert temporal correlation to color
            rgba_color = cm.plasma_r(norm(c3))
            # relation of b_dir which explains the error
            if abs(c2) > 0:
                exp_err = c2
            elif abs(c2) == 0:
                exp_err = 0
            # diagnose the error
            if abs(c1) <= l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(c1) > l and exp_err <= l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(c1) > l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            # FBM
            elif abs(c1) <= l and exp_err <= l and eff <= eff_l:
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(0, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(np.pi, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
            # FGM
            elif abs(c1) <= l and exp_err <= l and eff > eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff_min >= 0:
            ax.set_rmax(0)
        elif eff_min <= 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", "", "", ""])
        ax.text(
            -0.04,
            0.5,
            r"Comp2 < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"Comp2 > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"Comp1 < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"Comp1 > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("Comp3", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        return fig

    elif extended:
        fig = plt.figure(figsize=(12, 6), constrained_layout=True)
        gs = fig.add_gridspec(1, 2)
        ax = fig.add_subplot(gs[0, 0], projection="polar")
        ax1 = fig.add_axes([0.66, 0.3, 0.32, 0.32], frameon=True)
        # dummie plot for colorbar of temporal correlation
        cs = np.arange(0, 1.1, 0.1)
        dummie_cax = ax.scatter(cs, cs, c=cs, cmap="plasma_r")
        # Clear axis
        ax.cla()
        # plot regions
        ax.plot(
            (1, np.deg2rad(45)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(135)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(225)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        ax.plot(
            (1, np.deg2rad(315)),
            (1, np.min(yy)),
            color="lightgray",
            linewidth=1.5,
            ls="--",
            zorder=0,
        )
        # contours of DE
        cp = ax.contour(theta, r, r, colors="darkgray", levels=c_levels, zorder=1)
        cl = ax.clabel(cp, inline=True, fontsize=10, fmt="%1.1f", colors="dimgrey")
        # threshold efficiency for FBM
        eff_l = 1 - np.sqrt((l) ** 2 + (l) ** 2 + (l) ** 2)
        # loop over each data point
        for (c1, c2, c3, eff) in zip(ll_comp1, ll_comp2, ll_comp3, ll_eff):
            # normalize threshold with mean flow becnhmark
            ang = np.arctan2(c1, c2)
            # convert temporal correlation to color
            rgba_color = cm.plasma_r(norm(c3))
            # relation of b_dir which explains the error
            if abs(c2) > 0:
                exp_err = c2
            elif abs(c2) == 0:
                exp_err = 0
            # diagnose the error
            if abs(c1) <= l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(c1) > l and exp_err <= l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            elif abs(c1) > l and exp_err > l and eff <= eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
            # FBM
            elif abs(c1) <= l and exp_err <= l and eff <= eff_l:
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(0, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
                ax.annotate(
                    "",
                    xytext=(0, 1),
                    xy=(np.pi, eff),
                    arrowprops=dict(facecolor=rgba_color),
                )
            # FGM
            elif abs(c1) <= l and exp_err <= l and eff > eff_l:
                c = ax.scatter(ang, eff, color=rgba_color, zorder=2)
        ax.set_rticks([])  # turn default ticks off
        ax.set_rmin(1)
        if eff_min >= 0:
            ax.set_rmax(0)
        elif eff_min < 0:
            ax.set_rmax(-ax_lim + 0.2)
        # turn labels and grid off
        ax.tick_params(
            labelleft=False,
            labelright=False,
            labeltop=False,
            labelbottom=True,
            grid_alpha=0.01,
        )
        ax.set_xticklabels(["", "", "", "", "", "", "", ""])
        ax.text(
            -0.04,
            0.5,
            r"Comp2 < 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            1.04,
            0.5,
            r"Comp2 > 0",
            va="center",
            ha="center",
            rotation=90,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            -0.04,
            r"Comp1 < 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        ax.text(
            0.5,
            1.04,
            r"Comp1 > 0",
            va="center",
            ha="center",
            rotation=0,
            rotation_mode="anchor",
            transform=ax.transAxes,
        )
        # add colorbar for temporal correlation
        cbar = fig.colorbar(
            dummie_cax, ax=ax, orientation="horizontal", ticks=[1, 0.5, 0], shrink=0.8
        )
        cbar.set_label("Comp3", labelpad=4)
        cbar.set_ticklabels(["1", "0.5", "<0"])
        cbar.ax.tick_params(direction="in")

        # convert to degrees
        phi = np.arctan2(comp1, comp2)
        diag_deg = (phi * (180 / np.pi)) + 135
        diag_deg[diag_deg < 0] = 360 - diag_deg[diag_deg < 0]

        # 1-D density plot
        g = sns.kdeplot(diag_deg, color="k", ax=ax1)
        ax1.set_xticks([0, 90, 180, 270, 360])
        ax1.set_xlim(0, 360)
        ax1.set_ylim(0,)
        ax1.set(ylabel="Density", xlabel=r"[$^\circ$]")

        # 2-D density plot
        c3_colors = cm.plasma_r(norm(comp3))
        g = sns.jointplot(
            diag_deg,
            eff,
            kind="kde",
            zorder=1,
            n_levels=20,
            cmap="Greys",
            shade_lowest=False,
            marginal_kws={"color": "k", "shade": False},
        ).plot_joint(plt.scatter, c=c3_colors, alpha=0.4, zorder=2)
        g.set_axis_labels(r"[$^\circ$]", r"Eff [-]")
        g.ax_joint.set_xticks([0, 90, 180, 270, 360])
        g.ax_joint.set_xlim(0, 360)
        g.ax_joint.set_ylim(-ax_lim, 1)
        g.ax_marg_x.set_xticks([0, 90, 180, 270, 360])
        kde_data = g.ax_marg_y.get_lines()[0].get_data()
        kde_xx = kde_data[0]
        kde_yy = kde_data[1]
        norm = matplotlib.colors.Normalize(vmin=-ax_lim, vmax=1.0)
        colors = cm.Reds_r(norm(kde_yy))
        npts = len(kde_xx)
        for i in range(npts - 1):
            g.ax_marg_y.fill_betweenx(
                [kde_yy[i], kde_yy[i + 1]], [kde_xx[i], kde_xx[i + 1]], color=colors[i]
            )
        g.fig.tight_layout()

        return fig, g.fig



# -*- coding: utf-8 -*-

"""
de.misc
~~~~~~~~~~~

:2019 by Robin Schwemmle.
:license: GNU GPLv3, see LICENSE for more details.
"""

import datetime as dt
import pandas as pd
import numpy as np
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import scipy as sp
import seaborn as sns


def sort_obs(ts):
    """
    Sort time series by observed values.

    Parameters
    ----------
    ts : dataframe
        Dataframe with two time series (e.g. observed and simulated)

    Returns
    ----------
    obs_sort : dataframe
        dataframe with two time series sorted by the observed values
        in ascending order
    """
    df_ts = pd.DataFrame(data=ts)
    obs_sort = df_ts.sort_values(by=['Qobs'], ascending=False)

    return obs_sort

def fdc_obs_sort(Q):
    """
    Plotting the flow duration curves of observed and simulated runoff.
    Descending order of ebserved time series is applied to simulated time
    series.

    Parameters
    ----------
    Q : dataframe
        Containing time series of Qobs and Qsim.
    """
    df_Q = pd.DataFrame(data=Q)
    df_Q_sort = sort_obs(df_Q)

    # calculate exceedence probability
    ranks_obs = sp.stats.rankdata(df_Q_sort['Qobs'], method='ordinal')
    ranks_obs = ranks_obs[::-1]
    prob_obs = [(ranks_obs[i]/(len(df_Q_sort['Qobs'])+1)) for i in range(len(df_Q_sort['Qobs']))]

    fig, ax = plt.subplots()
    ax.plot(prob_obs, df_Q_sort['Qsim'], color='red', label='Simulated')
    ax.plot(prob_obs, df_Q_sort['Qobs'], color='blue', label='Observed')
    ax.set(ylabel=_q_lab,
           xlabel='Exceedence probabilty [-]', yscale='log')
    ax.legend(loc=1)

def calc_de_sort(obs, sim):
    """
    Calculate Diagnostic-Efficiency (DE).

    Simulated values are sorted by order of observed values.

    Parameters
    ----------
    obs : (N,)array_like
        Observed time series as 1-D array

    sim : (N,)array_like
        Simulated time series

    Returns
    ----------
    sig : float
        diagnostic efficiency measure
    """
    # mean relative bias
    brel_mean = de.calc_brel_mean(obs, sim)
    # remaining relative bias
    brel_rest = de.calc_brel_rest(obs, sim)
    # area of relative remaing bias
    b_area = de.calc_bias_area(brel_rest)
    # diagnostic efficiency
    sig = 1 - np.sqrt((brel_mean)**2 + (b_area)**2)

    return sig

def smooth_ts(ts, win=5):
    """
    Underestimate high flows - Overestimate low flows.

    Time series is smoothed by rolling average. Maxima decrease and minima
    decrease.

    Parameters
    ----------
    ts : dataframe
        Time series

    win : int, optional
        Size of window used to apply rolling mean. The default is 5 days.

    Returns
    ----------
    smoothed_ts : series
        Smoothed time series
    """
    smoothed_ts = ts.rolling(window=win).mean()
    smoothed_ts.fillna(method='bfill', inplace=True)

    return smoothed_ts

def _datacheck_peakdetect(x_axis, y_axis):
    """
    Check input data for peak detection.

    Parameters
    ----------
    x_axis : str
        path to file with meta informations on the catchments

    y_axis : str, optional
        delimeter to use. The default is ','.

    Returns
    ----------
    x_axis : dataframe
        imported time series
    """

    if x_axis is None:
        x_axis = range(len(y_axis))

    if len(y_axis) != len(x_axis):
        raise ValueError("Input vectors y_axis and x_axis must have same length")

    #needs to be a numpy array
    y_axis = np.array(y_axis)
    x_axis = np.array(x_axis)

    return x_axis, y_axis

def peakdetect(y_axis, x_axis = None, lookahead=200, delta=0):
    """
    Converted from/based on a MATLAB scripts at:
    http://billauer.co.il/peakdet.html
    https://gist.github.com/sixtenbe/1178136

    Function for detecting local maxima and minima in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maxima and minima respectively

    Parameters
    ----------
    y_axis : array_like
        contains the signal over which to find peaks

    x_axis : array_like, optional
        values correspond to the y_axis list and is used
        in the return to specify the position of the peaks. If omitted an
        index of the y_axis is used.
        (default: None)

    lookahead : int, optional
        distance to look ahead from a peak candidate to determine if
        it is the actual peak
        '(samples / period) / f' where '4 >= f >= 1.25' might be a good value

    delta : int, optional
        this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. It is recommended that delta should be set to
        delta >= RMSnoise * 5.
        When omitted delta function causes a 20% decrease in speed.
        When used correctly it can double the speed of the function.

    Returns
    ----------
    max_peaks : list
        containing the positive peaks. Each cell of the list contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)

    min_peaks : list
        containing the negative peaks. Each cell of the list contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    max_peaks = []
    min_peaks = []
    dump = []  # Used to pop the first hit which is false

    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # store data length for later use
    length = len(y_axis)

    # perform some checks
    if lookahead < 1:
        raise ValueError("Lookahead must be '1' or above in value")
    if not (np.isscalar(delta) and delta >= 0):
        raise ValueError("delta must be a positive number")

    # maxima and minima candidates are temporarily stored in
    # mx and mn respectively
    mn, mx = np.Inf, -np.Inf

    # Only detect peak if there is 'lookahead' amount of points after it
    for index, (x, y) in enumerate(zip(x_axis[:-lookahead], y_axis[:-lookahead])):
        if y > mx:
            mx = y
            mxpos = x
        if y < mn:
            mn = y
            mnpos = x

        #### look for max ####
        if y < mx-delta and mx != np.Inf:
            # Maxima peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].max() < mx:
                max_peaks.append([mxpos, mx])
                dump.append(True)
                #set algorithm to only find minima now
                mx = np.Inf
                mn = np.Inf
                if index+lookahead >= length:
                    # end is within lookahead no more peaks can be found
                    break
                continue
            #else:  #slows shit down this does
            #    mx = ahead
            #    mxpos = x_axis[np.where(y_axis[index:index+lookahead]==mx)]

        #### look for min ####
        if y > mn+delta and mn != -np.Inf:
            # Minima peak candidate found
            # look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].min() > mn:
                min_peaks.append([mnpos, mn])
                dump.append(False)
                # set algorithm to only find maxima now
                mn = -np.Inf
                mx = -np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
            #else:  #slows shit down this does
            #    mn = ahead
            #    mnpos = x_axis[np.where(y_axis[index:index+lookahead]==mn)]


    #Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
        del dump
    except IndexError:
        # no peaks were found, should the function return empty lists?
        pass

    return max_peaks, min_peaks

def disaggregate_obs(ts, max_peaks_ind, min_peaks_ind):
    """
    Overestimate high flows - Underestimate low flows.

    Increase max values and decrease min values.


    Parameters
    ----------
    ts : dataframe
        Dataframe with time series

    max_peaks_ind : list
        Index where max. peaks occur

    min_peaks_ind : list
        Index where min. peaks occur

    Returns
    ----------
    ts : dataframe
        Disaggregated time series
    """
    idx_max = max_peaks_ind - dt.timedelta(days=1)
    idxu_max = max_peaks_ind.union(idx_max)
    idx1_max = max_peaks_ind - dt.timedelta(days=2)

    idx_min = min_peaks_ind + dt.timedelta(days=1)
    idxu_min = min_peaks_ind.union(idx_min)
    idx1_min = min_peaks_ind + dt.timedelta(days=2)

    ts.loc[idx1_max, 'Qobs'] = ts.loc[idx1_max, 'Qobs'] * 1.1
    ts.loc[idx_max, 'Qobs'] = ts.loc[idx_max, 'Qobs'] * 1.2
    ts.loc[max_peaks_ind, 'Qobs'] = ts.loc[max_peaks_ind, 'Qobs'] * 1.3

    ts.loc[idx1_min, 'Qobs'] = ts.loc[idx1_min, 'Qobs'] * 0.5
    ts.loc[idx_min, 'Qobs'] = ts.loc[idx_min, 'Qobs'] * 0.5
    ts.loc[min_peaks_ind, 'Qobs'] = ts.loc[min_peaks_ind, 'Qobs'] * 0.5

    return ts

def plot_peaks(ts, max_peak_ts, min_peak_ts):
    """
    Plot time series.

    Parameters
    ----------
    ts : dataframe
        dataframe with time series

    max_peak_ts : dataframe
        dataframe with max. peaks of time series

    min_peak_ts : dataframe
        dataframe with min. peaks of time series
    """
    fig, ax = plt.subplots()
    ax.plot(ts.index, ts.iloc[:, 0].values, color='blue')
    ax.plot(max_peak_ts.index, max_peak_ts.iloc[:, 0].values, 'r.')
    ax.plot(min_peak_ts.index, min_peak_ts.iloc[:, 0].values, 'g.')
    ax.set(ylabel=r'[$m^{3}$ $s^{-1}$]',
           xlabel='Time [Days]')

def linear_bm(ts):
    """Generate a linear flow duration as a benchmark.

    Parameters
    ----------
    ts : dataframe
        Observed time series

    Returns
    ----------
    ts_lin: dataframe
        Smoothed time series
    """
    obs_sim = pd.DataFrame(index=ts.index, columns=['Qobs', 'Qsim'])
    obs_sim.iloc[:, 0] = ts.iloc[:, 0]
    # sort values by descending order
    obs_sort = obs_sim.sort_values(by='Qobs', ascending=False)
    nn = len(obs_sim.index)
    qmax = np.max(obs_sim.iloc[:, 0].values)
    qmin = np.min(obs_sim.iloc[:, 0].values)
    # linearly interpolated array
    lin_arr = np.linspace(qmin, qmax, nn)
    obs_sort.iloc[:, 1] = lin_arr
    # sort by index
    obs_sim = obs_sort.sort_index()
    ts_lin = obs_sim.iloc[:, 1].copy().to_frame()

    return ts_lin



"""Testing kge module."""

import os
import sys

import pytest
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from de import kge

baseline_dir = "images"

baseline_dir_single = os.path.join(baseline_dir, "kge/single")
baseline_dir_multi = os.path.join(baseline_dir, "kge/multi")

WIN = sys.platform.startswith("win")

# In some cases, the fonts on Windows can be quite different
DEFAULT_TOLERANCE = 10 if WIN else 2


def test_kge_for_arrays():
    eff = kge.calc_kge(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert eff == pytest.approx(0.683901305466148, rel=1e-4)


def test_kge_skill_for_arrays():
    eff = kge.calc_kge_skill(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
        bench=np.array([1, 1.1, 1.15, 1.15, 1.1, 1]),
    )
    assert eff == pytest.approx(0.8467044616487865, rel=1e-4)


def test_beta_for_arrays():
    beta = kge.calc_kge_beta(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert beta == pytest.approx(1.0980392156862746, rel=1e-4)


def test_alpha_for_arrays():
    alpha = kge.calc_kge_alpha(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert alpha == pytest.approx(1.2812057455166919, rel=1e-4)


def test_gamma_for_arrays():
    gamma = kge.calc_kge_gamma(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert gamma == pytest.approx(1.166812375381273, rel=1e-4)


def test_temp_cor_for_arrays():
    temp_cor = kge.calc_temp_cor(
        obs=np.array([1.5, 1, 0.8, 0.85, 1.5, 2]),
        sim=np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5]),
    )
    assert temp_cor == pytest.approx(0.8940281850583509, rel=1e-4)


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_single, tolerance=DEFAULT_TOLERANCE
)
def test_single_diag_polar_plot():
    obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2])
    sim = np.array([1.6, 1.3, 1, 0.8, 1.2, 2.5])
    fig = kge.polar_plot(obs, sim)

    return fig


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_multi, tolerance=DEFAULT_TOLERANCE
)
def test_multi_diag_polar_plot():
    beta = np.array([1.1, 1.15, 1.2, 1.1, 1.05, 1.15])
    alpha = np.array([1.15, 1.1, 1.2, 1.1, 1.1, 1.2])
    r = np.array([0.9, 0.85, 0.8, 0.9, 0.85, 0.9])
    eff_kge = np.array([0.79, 0.77, 0.65, 0.83, 0.81, 0.73])
    fig = kge.polar_plot_multi(beta, alpha, r, eff_kge)

    return fig



"""Testing util module."""

import os
import sys

import pytest
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from de import util

baseline_dir = "images"

baseline_dir_single = os.path.join(baseline_dir, "util/single_ts")
baseline_dir_dual = os.path.join(baseline_dir, "util/dual_ts")

WIN = sys.platform.startswith("win")

# In some cases, the fonts on Windows can be quite different
DEFAULT_TOLERANCE = 10 if WIN else 2


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_single, tolerance=DEFAULT_TOLERANCE
)
def test_single_ts():
    date_rng = pd.date_range(start="1/1/2018", periods=11)
    arr = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    ts = pd.Series(data=arr, index=date_rng)
    fig = util.plot_ts(ts)

    return fig


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir_dual, tolerance=DEFAULT_TOLERANCE
)
def test_dual_ts():
    date_rng = pd.date_range(start="1/1/2018", periods=11)
    obs = np.array([1.5, 1, 0.8, 0.85, 1.5, 2, 2.5, 3.5, 1.8, 1.5, 1.2])
    ts_obs = pd.Series(data=obs, index=date_rng)
    sim = np.array([1.4, 0.9, 1, 0.95, 1.4, 2.1, 2.6, 3.6, 1.9, 1.4, 1.1])
    ts_sim = pd.Series(data=sim, index=date_rng)
    fig = util.plot_obs_sim(ts_obs, ts_sim)

    return fig



# -*- coding: utf-8 -*-

"""
DE - Diagnostic Efficiency
"""

from ._version import get_versions

__title__ = "de"
__author__ = "Robin Schwemmle"
__license__ = "GNU GPLv3"
__version__ = get_versions()["version"]
del get_versions



"""Setup for the diag-eff package."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import versioneer

long_description = """diag-eff is an open-source toolbox written in Python for
specific evaluation of model performance. The toolbox provides functions to
calculate the Diagnostic Efficiency metric and and fucntions to visualize
contribution of metric terms by diagnostic polar plots. Note that the data
management of time series is handled using pandas data frame objects.
"""

INSTALL_REQUIRES = [
    "numpy",
    "scipy",
    "matplotlib",
    "seaborn",
    "pandas",
]
TEST_REQUIRES = [
    # testing and coverage
    "pytest",
    "coverage",
    "pytest-cov",
    # to be able to run `python setup.py checkdocs`
    "collective.checkdocs",
    "pygments",
]

setup(
    name="diag-eff",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="diag-eff: Diagnostic Efficiency",
    long_description=long_description,
    url="https://github.com/schwemro/diag-eff",
    author="Robin Schwemmle",
    author_email="robin.schwemmle@hydrology.uni-freiburg.de",
    license="GPLv3",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Hydrology",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    python_requires=">=3.6",
    packages=find_packages(exclude=["docs"]),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require={"test": TEST_REQUIRES + INSTALL_REQUIRES, },
)
